@@ -3,7 +3,7 @@
 #
 # kluster/kox point process models
 #
-# $Revision: 1.188 $ $Date: 2021/10/28 04:26:26 $
+# $Revision: 1.201 $ $Date: 2022/02/11 03:05:49 $
 #
 
 
@@ -71,6 +71,7 @@ kppm.ppp <- kppm.quad <-
            improve.args = list(),
            weightfun=NULL,
            control=list(),
+           stabilize=TRUE,
            algorithm,
            statistic="K",
            statargs=list(),
@@ -98,12 +99,13 @@ kppm.ppp <- kppm.quad <-
                       improve.args = improve.args,
                       weightfun=weightfun,
                       control=control,
+                      stabilize=stabilize,
                       algorithm=algorithm,
                       statistic=statistic,
                       statargs=statargs,
                       rmax = rmax)
   Xenv <- list2env(as.list(covariates), parent=parent.frame())
-  X <- eval(substitute(X), envir=Xenv, enclos=baseenv())
+  X <- eval(substitute(X), envir=Xenv, enclos=parent.frame())
   isquad <- is.quad(X)
   if(!is.ppp(X) && !isquad)
     stop("X should be a point pattern (ppp) or quadrature scheme (quad)")
@@ -126,7 +128,8 @@ kppm.ppp <- kppm.quad <-
             covfunargs=covfunargs, use.gam=use.gam, nd=nd, eps=eps)
   XX <- if(isquad) X$data else X
   
-  ## set default weight function
+           
+  ## default weight function
   if(is.null(weightfun))
     switch(method,
            adapcl = {
@@ -146,14 +149,17 @@ kppm.ppp <- kppm.quad <-
   ## fit
   out <- switch(method,
          mincon = kppmMinCon(X=XX, Xname=Xname, po=po, clusters=clusters,
-                             control=control, statistic=statistic,
+                             control=control, stabilize=stabilize,
+                             statistic=statistic,
                              statargs=statargs, rmax=rmax,
                              algorithm=algorithm, ...),
          clik2   = kppmComLik(X=XX, Xname=Xname, po=po, clusters=clusters,
-                             control=control, weightfun=weightfun, 
+                             control=control, stabilize=stabilize,
+                             weightfun=weightfun, 
                              rmax=rmax, algorithm=algorithm, ...),
          palm   = kppmPalmLik(X=XX, Xname=Xname, po=po, clusters=clusters,
-                             control=control, weightfun=weightfun, 
+                             control=control, stabilize=stabilize,
+                             weightfun=weightfun, 
                              rmax=rmax, algorithm=algorithm, ...),
          adapcl   = kppmCLadap(X=XX, Xname=Xname, po=po, clusters=clusters,
                              control=control, epsilon=epsilon, 
@@ -181,8 +187,12 @@ kppm.ppp <- kppm.quad <-
   return(out)
 }
 
-kppmMinCon <- function(X, Xname, po, clusters, control, statistic, statargs,
-                       algorithm="Nelder-Mead", DPP=NULL, ...) {
+## >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+##                M  i  n  i  m  u  m       C  o  n  t  r  a  s  t
+## <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
+
+kppmMinCon <- function(X, Xname, po, clusters, control=list(), stabilize=TRUE, statistic, statargs,
+                       algorithm="Nelder-Mead", DPP=NULL, ..., pspace=NULL) {
   # Minimum contrast fit
   stationary <- is.stationary(po)
   # compute intensity
@@ -202,22 +212,23 @@ kppmMinCon <- function(X, Xname, po, clusters, control, statistic, statargs,
     po <- tmp$po
   }
   mcfit <- clusterfit(X, clusters, lambda = lambda,
-                      dataname = Xname, control = control,
+                      dataname = Xname, control = control,  stabilize=stabilize,
                       statistic = statistic, statargs = statargs,
-                      algorithm=algorithm, ...)
+                      algorithm=algorithm, pspace=pspace, ...)
   fitinfo <- attr(mcfit, "info")
   attr(mcfit, "info") <- NULL
   # all info that depends on the fitting method:
-  Fit <- list(method    = "mincon",
-              statistic = statistic,
-              Stat      = fitinfo$Stat,
-              StatFun   = fitinfo$StatFun,
-              StatName  = fitinfo$StatName,
-              FitFun    = fitinfo$FitFun,
-              statargs  = statargs,
-              pspace    = fitinfo$pspace,
-              mcfit     = mcfit,
-              maxlogcl  = NULL)
+  Fit <- list(method       = "mincon",
+              statistic    = statistic,
+              Stat         = fitinfo$Stat,
+              StatFun      = fitinfo$StatFun,
+              StatName     = fitinfo$StatName,
+              FitFun       = fitinfo$FitFun,
+              statargs     = statargs,
+              pspace.given = pspace,
+              pspace.used  = fitinfo$pspace.used, 
+              mcfit        = mcfit,
+              maxlogcl     = NULL)
   # results
   if(!is.null(DPP)){
     clusters <- update(clusters, as.list(mcfit$par))
@@ -389,13 +400,14 @@ clusterfit <- function(X, clusters, lambda = NULL, startpar = NULL,
                                   margs=dots$margs,
                                   model=dots$model,
                                   funaux=info$funaux,
-                                  pspace=pspace),
+                                  pspace=pspace), ## As modified above
                              list(...)
                              )
 
   if(isDPP && algorithm=="Brent" && changealgorithm)
     mcargs <- resolve.defaults(mcargs, list(lower=alg$lower, upper=alg$upper))
 
+  
   ## .............. FIT .......................
   if(verbose) splat("Starting minimum contrast fit")
   mcfit <- do.call(mincontrast, mcargs)
@@ -444,22 +456,26 @@ clusterfit <- function(X, clusters, lambda = NULL, startpar = NULL,
   ## The old fit fun that would have been used (DO WE NEED THIS?)
   FitFun <- paste0(tolower(clusters), ".est", statistic)
 
-  extra <- list(FitFun    = FitFun,
-                Stat      = Stat,
-                StatFun   = StatFun,
-                StatName  = StatName,
-                modelname  = info$modelabbrev,
-                isPCP      = isPCP,
-                lambda     = lambda,
-                pspace     = pspace)
+  extra <- list(FitFun       = FitFun,
+                Stat         = Stat,
+                StatFun      = StatFun,
+                StatName     = StatName,
+                modelname    = info$modelabbrev,
+                isPCP        = isPCP,
+                lambda       = lambda,
+                pspace.used  = pspace) # Modified from call to 'clusterfit'
   attr(mcfit, "info") <- extra
   if(verbose) splat("Returning from clusterfit")
   return(mcfit)
 }
 
+## >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+##                C  o  m  p  o  s  i  t  e    L  i  k  e  l  i  h  o  o  d
+## <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
 
-kppmComLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
-                       algorithm="Nelder-Mead", DPP=NULL, ..., pspace=NULL) {
+kppmComLik <- function(X, Xname, po, clusters, control=list(), stabilize=TRUE,
+                       weightfun, rmax, algorithm="Nelder-Mead",
+                       DPP=NULL, ..., pspace=NULL) {
   W <- as.owin(X)
   if(is.null(rmax))
     rmax <- rmax.rule("K", W, intensity(X))
@@ -471,7 +487,7 @@ kppmComLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
   # compute weights for pairs of points
   if(is.function(weightfun)) {
     wIJ <- weightfun(dIJ)
-    sumweight <- safevalue(sum(wIJ))
+    sumweight <- safePositiveValue(sum(wIJ))
   } else {
     npairs <- length(dIJ)
     wIJ <- rep.int(1, npairs)
@@ -507,7 +523,7 @@ kppmComLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
     # with density proportional to intensity function
     g <- distcdf(M, dW=lambdaM, delta=rmax/4096)
     # scaling constant is (integral of intensity)^2
-    gscale <- safevalue(integral.im(lambdaM)^2, default=npoints(X)^2)
+    gscale <- safePositiveValue(integral.im(lambdaM)^2, default=npoints(X)^2)
   }
 
   # Detect DPP model and change clusters and intensity correspondingly
@@ -562,16 +578,16 @@ kppmComLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
     #       sum(log(lambdaIJ)) - npairs * log(gscale)
     obj <- function(par, objargs) {
       with(objargs, {
-        logprod <- sum(log(safevalue(paco(dIJ, par))))
+        logprod <- sum(log(safePositiveValue(paco(dIJ, par))))
         integ <- unlist(stieltjes(paco, g, par=par))
         integ <- pmax(SMALLVALUE, integ)
         logcl <- 2*(logprod - sumweight * log(integ))
-        logcl <- safevalue(logcl, default=-BIGVALUE)
+        logcl <- safeFiniteValue(logcl, default=-BIGVALUE)
         return(logcl)
       },
       enclos=objargs$envir)
     }
-    ## determine a suitable large number to replace Inf
+    ## Determine a suitable large number to replace Inf
     objargs$BIGVALUE <- bigvaluerule(obj, objargs, startpar)
   } else {
     # create local function to evaluate  pair correlation(d) * weight(d)
@@ -595,22 +611,41 @@ kppmComLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
       {
         integ <- unlist(stieltjes(wpaco, g, par=par))
         integ <- pmax(SMALLVALUE, integ)
-        logcl <- safevalue(
-          2*(sum(wIJ * log(safevalue(paco(dIJ, par))))
+        logcl <- safeFiniteValue(
+          2*(sum(wIJ * log(safePositiveValue(paco(dIJ, par))))
             - sumweight * log(integ)),
           default=-BIGVALUE)
         return(logcl)
       },
       enclos=objargs$envir)
     }
-    ## determine a suitable large number to replace Inf
+    ## Determine a suitable large number to replace Inf
     objargs$BIGVALUE <- bigvaluerule(obj, objargs, startpar)
   }
-  #' .........................................................
-  ## arguments for optimization
-  ctrl <- resolve.defaults(list(fnscale=-1), control, list(trace=0))
-  optargs <- list(par=startpar, fn=obj, objargs=objargs, control=ctrl, method=algorithm)
-  ## DPP resolving algorithm and checking startpar
+
+  ## ......................  Optimization settings  ........................
+
+  if(stabilize) {
+    ## Numerical stabilisation 
+    ## evaluate objective at starting state
+    startval <- obj(startpar, objargs)
+    ## use to determine appropriate global scale
+    smallscale <- sqrt(.Machine$double.eps)
+    fnscale <- -max(abs(startval), smallscale)
+    parscale <- pmax(abs(startpar), smallscale)
+    scaling <- list(fnscale=fnscale, parscale=parscale)
+  } else {
+    scaling <- list(fnscale=-1)
+  }
+
+  ## Update list of algorithm control arguments
+  control.updated <- resolve.defaults(control, scaling, list(trace=0))
+
+  ## Initialise list of all arguments to 'optim'
+  optargs <- list(par=startpar, fn=obj, objargs=objargs,
+                  control=control.updated, method=algorithm)
+
+  ## DPP case: check startpar and modify algorithm
   changealgorithm <- length(startpar)==1 && algorithm=="Nelder-Mead"
   if(isDPP){
     alg <- dppmFixAlgorithm(algorithm, changealgorithm, clusters,
@@ -621,6 +656,8 @@ kppmComLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
       optargs$upper <- alg$upper
     }
   }
+  
+  
   ## ..........   optimize it ..............................
   opt <- do.call(optim, optargs)
 
@@ -637,14 +674,15 @@ kppmComLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
   ## Finish in DPP case
   if(!is.null(DPP)){
     ## all info that depends on the fitting method:
-    Fit <- list(method    = "clik2",
-                clfit     = opt,
-                weightfun = weightfun,
-                rmax      = rmax,
-                objfun    = obj,
-                objargs   = objargs,
-                maxlogcl  = opt$value,
-                pspace    = pspace)
+    Fit <- list(method       = "clik2",
+                clfit        = opt,
+                weightfun    = weightfun,
+                rmax         = rmax,
+                objfun       = obj,
+                objargs      = objargs,
+                maxlogcl     = opt$value,
+                pspace.given = pspace,
+                pspace.used  = pspace)
     # pack up
     clusters <- update(clusters, as.list(opt$par))
     result <- list(Xname      = Xname,
@@ -674,14 +712,15 @@ kppmComLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
           eval.im(log(lambda) - sigma2/2)    
   }
   # all info that depends on the fitting method:
-  Fit <- list(method    = "clik2",
-              clfit     = opt,
-              weightfun = weightfun,
-              rmax      = rmax,
-              objfun    = obj,
-              objargs   = objargs,
-              maxlogcl  = opt$value,
-              pspace    = pspace)
+  Fit <- list(method       = "clik2",
+              clfit        = opt,
+              weightfun    = weightfun,
+              rmax         = rmax,
+              objfun       = obj,
+              objargs      = objargs,
+              maxlogcl     = opt$value,
+              pspace.given = pspace,
+              pspace.used  = pspace)
   # pack up
   result <- list(Xname      = Xname,
                  X          = X,
@@ -704,7 +743,12 @@ kppmComLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
 }
 
 
-kppmPalmLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
+## >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+##                P  a  l  m     L  i  k  e  l  i  h  o  o  d              
+## <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
+
+
+kppmPalmLik <- function(X, Xname, po, clusters, control=list(), stabilize=TRUE, weightfun, rmax,
                         algorithm="Nelder-Mead", DPP=NULL, ..., pspace=NULL) {
   W <- as.owin(X)
   if(is.null(rmax))
@@ -753,7 +797,7 @@ kppmPalmLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
     # and a random point in W with density proportional to intensity function
     g <- distcdf(X, M, dV=lambdaM, delta=rmax/4096)
     # scaling constant is (integral of intensity) * (number of points)
-    gscale <- safevalue(integral.im(lambdaM) * npoints(X),
+    gscale <- safePositiveValue(integral.im(lambdaM) * npoints(X),
                         default=npoints(X)^2)
   }
 
@@ -801,7 +845,7 @@ kppmPalmLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
   if(!is.function(weightfun)) {
     # pack up necessary information
     objargs <- list(dIJ=dIJ, g=g, gscale=gscale,
-                    sumloglam=safevalue(sum(log(lambdaJ))),
+                    sumloglam=safeFiniteValue(sum(log(lambdaJ))),
                     envir=environment(paco),
                     BIGVALUE=1, # updated below
                     SMALLVALUE=.Machine$double.eps)
@@ -811,14 +855,15 @@ kppmPalmLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
       with(objargs, {
         integ <- unlist(stieltjes(paco, g, par=par))
         integ <- pmax(SMALLVALUE, integ)
-        logplik <- safevalue(sumloglam + sum(log(safevalue(paco(dIJ, par))))
-                           - gscale * integ,
-                           default=-BIGVALUE)
+        logplik <- safeFiniteValue(
+                sumloglam + sum(log(safePositiveValue(paco(dIJ, par))))
+                - gscale * integ,
+          default=-BIGVALUE)
         return(logplik)
       },
       enclos=objargs$envir)
     }
-    ## determine a suitable large number to replace Inf
+    ## Determine a suitable large number to replace Inf
     objargs$BIGVALUE <- bigvaluerule(obj, objargs, startpar)
   } else {
     # create local function to evaluate  pair correlation(d) * weight(d)
@@ -831,7 +876,9 @@ kppmPalmLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
     }
     # pack up necessary information
     objargs <- list(dIJ=dIJ, wIJ=wIJ, g=g, gscale=gscale,
-                    wsumloglam=safevalue(sum(wIJ * safevalue(log(lambdaJ)))),
+                    wsumloglam=safeFiniteValue(
+                      sum(wIJ * safeFiniteValue(log(lambdaJ)))
+                    ),
                     envir=environment(wpaco),
                     BIGVALUE=1, # updated below
                     SMALLVALUE=.Machine$double.eps)
@@ -841,22 +888,41 @@ kppmPalmLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
       with(objargs, {
         integ <- unlist(stieltjes(wpaco, g, par=par))
         integ <- pmax(SMALLVALUE, integ)
-        logplik <- safevalue(wsumloglam +
-                             sum(wIJ * log(safevalue(paco(dIJ, par))))
+        logplik <- safeFiniteValue(wsumloglam +
+                             sum(wIJ * log(safePositiveValue(paco(dIJ, par))))
                              - gscale * integ,
                              default=-BIGVALUE)
         return(logplik)
       },
       enclos=objargs$envir)
     }
-    ## determine a suitable large number to replace Inf
+    ## Determine a suitable large number to replace Inf
     objargs$BIGVALUE <- bigvaluerule(obj, objargs, startpar)
   }
-  # arguments for optimization
-  ctrl <- resolve.defaults(list(fnscale=-1), control, list(trace=0))
+
+  ## ......................  Optimization settings  ........................
+
+  if(stabilize) {
+    ## Numerical stabilisation 
+    ## evaluate objective at starting state
+    startval <- obj(startpar, objargs)
+    ## use to determine appropriate global scale
+    smallscale <- sqrt(.Machine$double.eps)
+    fnscale <- -max(abs(startval), smallscale)
+    parscale <- pmax(abs(startpar), smallscale)
+    scaling <- list(fnscale=fnscale, parscale=parscale)
+  } else {
+    scaling <- list(fnscale=-1)
+  }
+
+  ## Update list of algorithm control arguments
+  control.updated <- resolve.defaults(control, scaling, list(trace=0))
+
+  ## Initialise list of all arguments to 'optim'
   optargs <- list(par=startpar, fn=obj, objargs=objargs,
-                  control=ctrl, method=algorithm)
-  ## DPP resolving algorithm and checking startpar
+                  control=control.updated, method=algorithm)
+
+  ## DPP case: check startpar and modify algorithm
   changealgorithm <- length(startpar)==1 && algorithm=="Nelder-Mead"
   if(isDPP){
     alg <- dppmFixAlgorithm(algorithm, changealgorithm, clusters,
@@ -867,6 +933,10 @@ kppmPalmLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
       optargs$upper <- alg$upper
     }
   }
+
+  ## .......................................................................
+  
+
   # optimize it
   opt <- do.call(optim, optargs)
   # raise warning/error if something went wrong
@@ -882,14 +952,15 @@ kppmPalmLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
   ## Finish in DPP case
   if(!is.null(DPP)){
     ## all info that depends on the fitting method:
-    Fit <- list(method    = "palm",
-                clfit     = opt,
-                weightfun = weightfun,
-                rmax      = rmax,
-                objfun    = obj,
-                objargs   = objargs,
-                maxlogcl  = opt$value,
-                pspace    = pspace)
+    Fit <- list(method       = "palm",
+                clfit        = opt,
+                weightfun    = weightfun,
+                rmax         = rmax,
+                objfun       = obj,
+                objargs      = objargs,
+                maxlogcl     = opt$value,
+                pspace.given = pspace,
+                pspace.used  = pspace)
     # pack up
     clusters <- update(clusters, as.list(optpar))
     result <- list(Xname      = Xname,
@@ -918,14 +989,15 @@ kppmPalmLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
           eval.im(log(lambda) - sigma2/2)    
   }
   # all info that depends on the fitting method:
-  Fit <- list(method    = "palm",
-              clfit     = opt,
-              weightfun = weightfun,
-              rmax      = rmax,
-              objfun    = obj,
-              objargs   = objargs,
-              maxlogcl  = opt$value,
-              pspace    = pspace)
+  Fit <- list(method       = "palm",
+              clfit        = opt,
+              weightfun    = weightfun,
+              rmax         = rmax,
+              objfun       = obj,
+              objargs      = objargs,
+              maxlogcl     = opt$value,
+              pspace.given = pspace,
+              pspace.used  = pspace)
   # pack up
   result <- list(Xname      = Xname,
                  X          = X,
@@ -946,6 +1018,11 @@ kppmPalmLik <- function(X, Xname, po, clusters, control, weightfun, rmax,
 }
 
 
+## >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+##       A  d  a  p  t  i  v  e     C  o  m  p  o  s  i  t  e    L  i  k  e  l  i  h  o  o  d
+## <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
+
+
 ## ........... contributed by Chiara Fend ...................
 ## needs nonlinear equation solver nleqslv
 
@@ -993,7 +1070,7 @@ kppmCLadap <- function(X, Xname, po, clusters, control, weightfun,
     # with density proportional to intensity function
     g <- distcdf(M, dW=lambdaM, delta=rmax/4096)
     # scaling constant is (integral of intensity)^2
-    gscale <- safevalue(integral.im(lambdaM)^2, default=npoints(X)^2)
+    gscale <- safePositiveValue(integral.im(lambdaM)^2, default=npoints(X)^2)
   }
   
   isDPP <- !is.null(DPP)
@@ -1631,18 +1708,20 @@ update.kppm <- function(object, ..., evaluate=TRUE, envir=environment(terms(obje
   methodname <- as.character(thecall[[1L]])
   switch(methodname,
          kppm.formula = {
-	   # original call has X = [formula with lhs]
+	   ## original call has X = [formula with lhs]
 	   if(!is.null(Xexpr)) {
 	     lhs.of.formula(fmla) <- Xexpr
 	   } else if(is.null(lhs.of.formula(fmla))) {
 	     lhs.of.formula(fmla) <- as.name('.')
 	   }
-           oldformula <- as.formula(getCall(object)$X)
+           oldformula <- getCall(object)$X
+           oldformula <- eval(oldformula, callframe)
            thecall$X <- newformula(oldformula, fmla, callframe, envir)
          },
          {
-	   # original call has X = ppp and trend = [formula without lhs]
-           oldformula <- as.formula(getCall(object)$trend %orifnull% (~1))
+	   ## original call has X = ppp and trend = [formula without lhs]
+           oldformula <- getCall(object)$trend %orifnull% (~1)
+           oldformula <- eval(oldformula, callframe)
 	   fom <-  newformula(oldformula, fmla, callframe, envir)
 	   if(!is.null(Xexpr))
 	      lhs.of.formula(fom) <- Xexpr
@@ -1701,7 +1780,7 @@ unitname.kppm <- unitname.dppm <- function(x) {
 as.fv.kppm <- as.fv.dppm <- function(x) {
   if(x$Fit$method == "mincon")
     return(as.fv(x$Fit$mcfit))
-  gobs <- pcfinhom(x$X, lambda=x, correction="good", update=FALSE)
+  gobs <- if(is.stationary(x)) pcf(x$X, correction="good") else pcfinhom(x$X, lambda=x, correction="good", update=FALSE)
   gfit <- (pcfmodel(x))(gobs$r)
   g <- bind.fv(gobs,
                data.frame(fit=gfit), 
@@ -1847,3 +1926,20 @@ psib.kppm <- function(object) {
   return(p)
 }
 
+reach.kppm <- function(x, ..., epsilon) {
+  thresh <- if(missing(epsilon)) NULL else epsilon
+  if(x$isPCP) 
+    return(2 * clusterradius.kppm(x, ..., thresh=thresh))
+  ## use pair correlation
+  g <- pcfmodel(x)
+  ## find upper bound
+  if(is.null(thresh)) thresh <- 0.01
+  f <- function(r) { g(r) - 1 - thresh }
+  scal <- as.list(x$par)$scale %orifnull% 1
+  for(a in scal * 2^(0:10)) { if(f(a) < 0) break; }
+  if(f(a) > 0) return(Inf)
+  ## solve g(r) = 1 + epsilon
+  b <- uniroot(f, c(0, a))$root
+  return(b)
+}
+
