@@ -1,14 +1,16 @@
-#
-#  mincontrast.R
-#
-#  Functions for estimation by minimum contrast
-#
+c#'
+#'  mincontrast.R
+#'
+#'  Functions for estimation by minimum contrast
+#'
+#'  $Revision: 1.119 $ $Date: 2022/01/21 04:07:12 $
+#' 
 
 
 ##################  base ################################
 
-safevalue <- function(x, default=0) {
-  ## ensure x is finite and acceptable to C routines
+safePositiveValue <- function(x, default=.Machine$double.eps) {
+  ## ensure x is finite, positive, and acceptable to C routines
   ifelse(is.finite(x),
          pmin(.Machine$double.xmax,
               pmax(.Machine$double.eps,
@@ -16,13 +18,33 @@ safevalue <- function(x, default=0) {
          default)
 }
 
+safeFiniteValue <- function(x, default=0) {
+  ## ensure x is finite and acceptable to C routines
+  biggest <- .Machine$double.xmax
+  ifelse(is.finite(x),
+         pmin(biggest,
+              pmax(-biggest,
+                   x)),
+         default)
+}
+
 bigvaluerule <- function(objfun, objargs, startpar, ...) {
-  ## evaluate objective at starting parameter vector
+  ## Determine suitable large number to replace Inf values of objective
+  ## Evaluate objective at starting parameter vector
   startval <- do.call(objfun,  list(par=startpar, objargs=objargs, ...))
-  ## determine suitable large number to replace Inf values of objective
-  bigvalue <- min(max(1, 10 * abs(startval)),
-                  with(.Machine, sqrt(double.xmax) * double.eps))
-  return(bigvalue)
+  ## check 
+  with(.Machine, {
+    hugeval <- sqrt(double.xmax) * double.eps
+    if(abs(startval) > hugeval) {
+      warning(paste("Internal error: objective function returns huge value",
+                    paren(startval),
+                    "which may cause numerical problems"),
+              call.=FALSE)
+      return(sqrt(double.xmax))
+    }
+    bigvalue <- min(hugeval, max(sqrt(hugeval), 1024 * abs(startval)))
+    return(bigvalue)
+  })
 }
 
 mincontrast <- local({
@@ -41,7 +63,7 @@ mincontrast <- local({
       discrep <- (abs(theo^qq - obsq))^pp
       ## protect C code from weird values
       bigvalue <- BIGVALUE + sqrt(sum(par^2))
-      discrep <- safevalue(discrep, default=bigvalue)
+      discrep <- safePositiveValue(discrep, default=bigvalue)
       ## rescaled integral of discrepancy
       value <- mean(discrep)
       
@@ -58,9 +80,11 @@ mincontrast <- local({
                           explain=list(dataname=NULL,
                                        modelname=NULL, fname=NULL),
                           action.bad.values=c("warn", "stop", "silent"),
+                          control=list(), stabilize=TRUE, 
                           pspace=NULL) {
     verifyclass(observed, "fv")
     action.bad.values <- match.arg(action.bad.values)
+
     
     stopifnot(is.function(theoretical))
     if(!any("par" %in% names(formals(theoretical))))
@@ -107,17 +131,31 @@ mincontrast <- local({
     ## sanity clause
     if(!all(ok <- is.finite(obs))) {
       doomed <- !any(ok)
-      whinge <- paste(if(doomed) "All" else "Some",
-                      "values of the empirical function",
-                      sQuote(explain$fname),
-                      "were infinite, NA or NaN.")
-      if(doomed || action.bad.values == "stop")
-        stop(whinge, call.=FALSE)
-      ## trim each end of domain
-      ra <- range(which(ok))
-      iMIN <- ra[1]
-      iMAX <- ra[2]
-      success <- all(ok[iMIN:iMAX])
+      if(!doomed && all(ok[-1])) {
+        ## common case: all finite except the value for r=0
+        whinge <- paste("The value of the empirical function",
+                        sQuote(explain$fname),
+                        "for r=", rvals[1],
+                        "was", paste0(obs[1], "."))
+        if(action.bad.values == "stop")
+          stop(whinge, call.=FALSE)
+        iMIN <- 2
+        iMAX <- length(obs)
+        success <- TRUE
+      } else {
+        ## general case: some non-finite values
+        whinge <- paste(if(doomed) "All" else "Some",
+                        "values of the empirical function",
+                        sQuote(explain$fname),
+                        "were infinite, NA or NaN.")
+        if(doomed || action.bad.values == "stop")
+          stop(whinge, call.=FALSE)
+        ## trim each end of domain
+        ra <- range(which(ok))
+        iMIN <- ra[1]
+        iMAX <- ra[2]
+        success <- all(ok[iMIN:iMAX])
+      }
       if(!success) {
         ## Finite and non-finite values are interspersed;
         ## find the longest run of finite values
@@ -167,8 +205,25 @@ mincontrast <- local({
                                      startpar, ...)
     
 
-    ## .  .  .  .  O  P  T  I  M  I  Z  E  .  .  .  .
-    minimum <- optim(startpar, fn=contrast.objective, objargs=objargs, ...)
+    ## ................... optimization algorithm control parameters .......................
+    if(stabilize) {
+      ## Numerical stabilisation 
+      ## evaluate objective at starting state
+      startval <- contrast.objective(startpar, objargs, ...)
+      ## use to determine appropriate global scale
+      smallscale <- sqrt(.Machine$double.eps)
+      fnscale <- max(abs(startval), smallscale)
+      parscale <- pmax(abs(startpar), smallscale)
+      scaling <- list(fnscale=fnscale, parscale=parscale)
+    } else {
+      scaling <- list() 
+    }
+    control <- resolve.defaults(control, scaling, list(trace=0))
+
+    ## .....................................................................................
+    ## >>>>>>>>>>>>>>>>>  .  .  .  .  O  P  T  I  M  I  Z  E  .  .  .  .  <<<<<<<<<<<<<<<<<<
+    minimum <- optim(startpar, fn=contrast.objective, objargs=objargs, ..., control=control)
+    ## .....................................................................................
     
     ## if convergence failed, issue a warning 
     signalStatus(optimStatus(minimum), errors.only=TRUE)
@@ -188,7 +243,8 @@ mincontrast <- local({
                    startpar = startpar,
                    objfun   = contrast.objective,
                    objargs  = objargs,
-                   dotargs  = list(...))
+                   dotargs  = list(...),
+                   pspace   = pspace)
     class(result) <- c("minconfit", class(result))
     return(result)
   }
@@ -225,13 +281,19 @@ print.minconfit <- function(x, ...) {
             paste(tagvalue, collapse=", "))
     }
   }
-  if(!is.null(fu) && !is.null(da))
-    splat("Fitted by matching theoretical", fu, "function to", da)
-  else {
-    if(!is.null(fu))
+  if(!is.null(fu)) {
+    if(length(fu) > 1) {
+      ## compress names like c("K", "inhom") -> "K[inhom]"
+      fsub <- paste(fu[-1], collapse=",")
+      fu <- paste0(fu[1], paren(fsub, "["))
+    }
+    if(!is.null(da)) {
+      splat("Fitted by matching theoretical", fu, "function to", da)
+    } else {
       splat(" based on", fu)
-    if(!is.null(da))
-      splat(" fitted to", da)
+    }
+  } else if(!is.null(da)) {
+    splat(" fitted to", da)
   }
 
   if(waxlyrical('space', terselevel))
@@ -806,28 +868,6 @@ cauchy.estpcf <- function(X, startpar=c(kappa=1,scale=1),
   return(result)
 }
 
-## user-callable
-resolve.vargamma.shape <- function(..., nu.ker=NULL, nu.pcf=NULL, default = FALSE) {
-  if(is.null(nu.ker) && is.null(nu.pcf)){
-    if(!default)
-        stop("Must specify either nu.ker or nu.pcf", call.=FALSE)
-    nu.ker <- -1/4
-  }
-  if(!is.null(nu.ker) && !is.null(nu.pcf))
-    stop("Only one of nu.ker and nu.pcf should be specified",
-         call.=FALSE)
-  if(!is.null(nu.ker)) {
-    check.1.real(nu.ker)
-    stopifnot(nu.ker > -1/2)
-    nu.pcf <- 2 * nu.ker + 1
-  } else {
-    check.1.real(nu.pcf)
-    stopifnot(nu.pcf > 0)
-    nu.ker <- (nu.pcf - 1)/2
-  }
-  return(list(nu.ker=nu.ker, nu.pcf=nu.pcf))
-}
-
 vargamma.estK <- function(X, startpar=c(kappa=1,scale=1), nu = -1/4,
                           lambda=NULL, q=1/4, p=2, rmin=NULL, rmax=NULL,
                           ...) {
@@ -856,7 +896,8 @@ vargamma.estK <- function(X, startpar=c(kappa=1,scale=1), nu = -1/4,
   ## Catch old nu.ker/nu.pcf syntax and resolve nu-value.
   dots <- list(...)
   if(missing(nu)){
-      nu <- resolve.vargamma.shape(nu.ker=dots$nu.ker, nu.pcf=dots$nu.pcf, default = TRUE)$nu.ker
+    nu <- resolve.vargamma.shape(nu.ker=dots$nu.ker,
+                                 nu.pcf=dots$nu.pcf, default = TRUE)$nu.ker
   }
   check.1.real(nu)
   stopifnot(nu > -1/2)
